#!.venv/bin/python3
import pyAgrum as gum
import psycopg2 
import sys
import argparse

def is_var_indicator(v):
    return not is_var(v)

def is_var(v):
    return not v.lower().startswith('i')


parser = argparse.ArgumentParser()
parser.add_argument("BNFile", help="Bayesian network file")
parser.add_argument("DBSize", help="size of the database")
args = parser.parse_args()

# inputs
bn_file = args.BNFile
db_size = int(args.DBSize)
out_file = "mcc.csv"
table_prefix = "test"

# loading the BN file
bn = gum.loadBN(bn_file)
var_names = sorted(list(bn.names()))

# generating the CSV file
g=gum.BNDatabaseGenerator(bn)
g.drawSamples(db_size)
g.toCSV(out_file)

# PG connection
conn = psycopg2.connect(database="mcc", 
                        user='postgres', password='postgres',  
                        host='127.0.0.1', port='5432')
conn.autocommit = True
cursor = conn.cursor()


#### CREATION OF D #####

# deleting the existing table
deletion_sql = "DROP TABLE IF EXISTS {};".format(table_prefix)
cursor.execute(deletion_sql)

# creating the table to store the data without missing values
col_def = map(lambda name : "{} int NOT NULL".format(name), var_names)
creation_sql = "CREATE TABLE {}(block SERIAL PRIMARY KEY, {});".format(table_prefix, ",".join(col_def))
cursor.execute(creation_sql) 

# loading the CSV data to the table
with open(out_file, "r") as file:
    next(file) # skip the header's line
    cursor.copy_from(file, table_prefix, sep=",", null="", columns=[v.lower() for v in var_names])

missing_selects = []
for v in var_names:
    if is_var(v) and v[1:] not in var_names:
        missing_selects.append("case when i{}=1 then NULL ELSE {} END".format(v,v))
    elif is_var(v):
        missing_selects.append(v)

#### CREATION OF D STAR ##### 

missing_data_table = "{}_star".format(table_prefix)
deletion_sql = "DROP TABLE IF EXISTS {};".format(missing_data_table)
cursor.execute(deletion_sql)
missing_selects_text = ",".join(missing_selects)
missing_data_sql = "CREATE TABLE {} AS SELECT block, {}, md5(CAST(({}) AS text)) AS superblock FROM {}".format(missing_data_table, missing_selects_text, missing_selects_text, table_prefix)
print(missing_data_sql)
cursor.execute(missing_data_sql)

#### CREATION OF BID #####

# superblock definitions with full tuples
superblock_tmp_table = "{}_sb_tmp".format(table_prefix)
deletion_sql = "DROP TABLE IF EXISTS {};".format(superblock_tmp_table)
cursor.execute(deletion_sql)

vars = list(filter(is_var, var_names))
col_def = map(lambda name : "{} int NOT NULL".format(name), vars)
creation_sql = "CREATE TABLE {}(superblock text, {}, prob float NOT NULL);".format(superblock_tmp_table, ",".join(col_def))
cursor.execute(creation_sql) 

cursor.execute("SELECT DISTINCT {}, superblock FROM {}".format(",".join(vars), missing_data_table))
ie = gum.LazyPropagation(bn)

insert_sql = "INSERT INTO {}({}, superblock, prob) VALUES({},%s,%s)".format(superblock_tmp_table, ",".join(vars), ",".join(["%s" for _ in vars]))

def instantiation_to_list(inst, row, vars):
    l = []
    i = 0
    inst_vars = list(map(lambda v : v.name(), inst.variablesSequence())) if inst is not None else []
    for v in vars:
        if v in inst_vars:
            l.append(inst[v])
        else:
            l.append(row[i])
        i+=1
    return l

for row in cursor.fetchall():
    null_vars = []
    null_pos  = []
    i = 0
    for v in vars:
        if row[i] is None:
            null_vars.append(v)
            null_pos.append(i)
        else:
            ie.addEvidence(v, row[i])
        i+=1
    ie.addJointTarget(set(null_vars))
    if len(null_vars) > 0:
        potentiel = ie.jointPosterior(set(null_vars)) if len(vars) > len(null_vars) else ie.evidenceJointImpact(vars,{})
        for inst in potentiel.loopIn():
            inserted_row = instantiation_to_list(inst, row, vars)
            inserted_row.append(row[-1])
            inserted_row.append(potentiel.get(inst))
            cursor.execute(insert_sql, inserted_row)
    else:
        inserted_row = instantiation_to_list(None, row, vars)
        inserted_row.append(row[-1])
        inserted_row.append(1)
        cursor.execute(insert_sql, inserted_row)
    ie.eraseAllJointTargets()
    ie.eraseAllEvidence()

# tuples table creation
tuple_table = "{}_tuple".format(table_prefix)
deletion_sql = "DROP TABLE IF EXISTS {};".format(tuple_table)
cursor.execute(deletion_sql)

cursor.execute("CREATE TABLE {} AS SELECT DISTINCT ROW_NUMBER() OVER (ORDER BY (SELECT 1)) AS tuple, * FROM (SELECT DISTINCT {} FROM {}) as t".format(tuple_table, ",".join(vars), superblock_tmp_table))

# superblocks table creation with tuple identifiers
superblock_table = "{}_sb".format(table_prefix)
deletion_sql = "DROP TABLE IF EXISTS {};".format(superblock_table)
cursor.execute(deletion_sql)

sb_tuple_preds = []
for v in vars:
    pred = "sb.{} = t.{}".format(v, v)
    sb_tuple_preds.append(pred)

cursor.execute("CREATE TABLE {} AS SELECT sb.superblock, t.tuple, sb.prob FROM {} AS sb, {} AS t WHERE {}".format(superblock_table, superblock_tmp_table, tuple_table, " AND ".join(sb_tuple_preds)))
    
deletion_sql = "DROP TABLE IF EXISTS {};".format(superblock_tmp_table)
cursor.execute(deletion_sql)



### Computation of the MCC
tuple_label_table = "{}_tuple_label".format(table_prefix)
block_label_table = "{}_block_label".format(table_prefix)
equigraph_table = "{}_equigraph".format(table_prefix)
matching_table =  "{}_matching".format(table_prefix)

nb_tuples = 8

def tuple_joint_probability(id):
    return 2*id/(nb_tuples * (nb_tuples + 1))

def tuple_occurence_weight(id, occurence):
    return 2*db_size*tuple_joint_probability(id) - 2*occurence +1

                 
def initialize_labels():
    deletion_sql = "DROP TABLE IF EXISTS {};".format(tuple_label_table)
    cursor.execute(deletion_sql)
    cursor.execute("CREATE TABLE {} (tuple INTEGER, occurence INTEGER, label float, weight float, PRIMARY KEY (tuple, occurence)) ".format(tuple_label_table))

    insert_sql = "INSERT INTO {}(tuple, occurence, label, weight) VALUES(%s,%s,%s,%s)".format(tuple_label_table)
    cursor.execute("SELECT tuple FROM {}".format(tuple_table))

    for row in cursor.fetchall():
        inserted_row = []
        tuple_id = row[0]
        inserted_row.append(tuple_id)
        inserted_row.append(1)
        inserted_row.append(0)
        inserted_row.append(tuple_occurence_weight(tuple_id, 1))
        cursor.execute(insert_sql, inserted_row)
        
    deletion_sql = "DROP TABLE IF EXISTS {};".format(block_label_table)
    cursor.execute(deletion_sql)
    cursor.execute("CREATE TABLE {} AS SELECT m.block, MAX(tl.weight) as label from {} as m, {} as s, {} as tl where m.superblock = s.superblock and s.tuple = tl.tuple GROUP BY m.block; ".format(block_label_table, missing_data_table, superblock_table, tuple_label_table))

def initialize_equigraph():
    deletion_sql = "DROP TABLE IF EXISTS {};".format(equigraph_table)
    cursor.execute(deletion_sql)
    cursor.execute("CREATE TABLE {} (block INTEGER, tuple INTEGER, max_occurence INTEGER) ".format(equigraph_table))
    cursor.execute("INSERT INTO {}(block, tuple, max_occurence)  SELECT m.block, s.tuple, 0 AS max_occurence FROM {} as m, {} as s where m.superblock = s.superblock".format(equigraph_table, missing_data_table, superblock_table))
    cursor.execute("UPDATE {} SET max_occurence = 1 WHERE (tuple, block) IN (SELECT tl.tuple, bl.block FROM {} AS tl, {} AS bl WHERE tl.weight = bl.label);".format(equigraph_table, tuple_label_table, block_label_table))

def initialize_matching():
    deletion_sql = "DROP TABLE IF EXISTS {};".format(matching_table)
    cursor.execute(deletion_sql)
    cursor.execute("CREATE TABLE {} (block INTEGER, tuple INTEGER, occurence INTEGER) ".format(matching_table))

# return a block free vertex in the equigraph, if it exists
def pick_free_block():
    cursor.execute("SELECT block FROM {} WHERE max_occurence > 0 AND block NOT IN (SELECT block FROM {})".format(equigraph_table, matching_table))
    res = cursor.fetchone()
    return res[0] if res is not None else None

# return the (tuple, occ) vertices that are the neighbors in the equigraph of at least one block in a set
def get_neighbors(blocks):
    neighbors = set()
    cursor.execute("SELECT tuple, max_occurence FROM {} WHERE max_occurence > 0 AND block IN %s".format(equigraph_table), [tuple(blocks)])
    for row in cursor.fetchall():
        neighbors.add(row)
    return neighbors

def get_matching_block(tuple, occurence):
    cursor.execute("SELECT block FROM {} WHERE tuple = {} AND occurence = {}".format(matching_table, tuple, occurence))
    res = cursor.fetchone()
    return res[0] if res is not None else None

def augmenting_path_insertion(u, y, prec):
    is_edge_matched = False
    while y != u:
        if is_edge_matched :
            cursor.execute("DELETE FROM {} WHERE block = %s AND tuple = %s AND occurence = %s".format(matching_table), [y, prec[y][0], prec[y][1]])
        else:
            cursor.execute("INSERT INTO {} (block, tuple, occurence) VALUES (%s, %s, %s)".format(matching_table), [prec[y], y[0], y[1]])
        y = prec[y]
        is_edge_matched = not is_edge_matched

def update_labels(S, T):
    cursor.execute("""
    SELECT e.block, e.tuple, e.max_occurence + 1, tl.weight - 2, tb.label, tl.label
    FROM {} AS e, {} AS tl, {} AS tb
    WHERE e.block IN %s AND e.max_occurence < %s AND e.max_occurence != 0 AND tl.tuple = e.tuple
    AND tl.occurence = e.max_occurence AND tb.block = e.block
    UNION ALL
    SELECT e.block, e.tuple, 1, tl.weight, tb.label, tl.label
    FROM {} AS e, {} AS tl, {} AS tb
    WHERE e.block IN %s AND e.max_occurence = 0 AND tl.tuple = e.tuple
    AND tl.occurence = 1 AND tb.block = e.block""".format(equigraph_table, tuple_label_table, block_label_table, equigraph_table, tuple_label_table, block_label_table), [tuple(S), db_size, tuple(S)])
    alpha = sys.float_info.max
    edges_to_add = []
    for row in cursor.fetchall():
        a = row[4] + row[5] - row[3]
        if a < alpha :
            alpha = a
            edges_to_add = []
        if a == alpha:
            edges_to_add.append(row)
    # print(alpha, edges_to_add)

    cursor.execute("UPDATE {} SET label = label - %s WHERE block in %s".format(block_label_table), [alpha, tuple(S)])
    cursor.execute("UPDATE {} SET label = label + %s WHERE (tuple, occurence) in %s".format(tuple_label_table), [alpha, tuple(T)])
    for edge in edges_to_add:
        cursor.execute("INSERT INTO {} (block, label) VALUES (%s, %s)".format(block_label_table), [row[0], row[4]])
        cursor.execute("INSERT  INTO {} (tuple, occurence, label, weight) VALUES (%s, %s, %s, %s) ON CONFLICT DO NOTHING".format(tuple_label_table), [row[1], row[2], row[5], row[3]])
        cursor.execute("UPDATE {} SET max_occurence = %s WHERE block = %s AND tuple = %s".format(equigraph_table), [row[2], row[0], row[1]])

def hungarian_step(u, S, T, S_neigh, prec):
    update_count = 0
    prec_z = u
    while True :
        remaining_neigh = S_neigh - T
        if len(remaining_neigh) > 0:
            y = next(iter(remaining_neigh))
            prec[y] = prec_z
            z = get_matching_block(y[0], y[1])
            if z is None :
                augmenting_path_insertion(u, y, prec)
                print(update_count)
                return
            else:
                S.add(z)
                T.add(y)
                prec[z] = y
                S_neigh.update(get_neighbors({z}))
        else:
            update_count+=1
            # print("update labels")
            # print(v, S, T, S_neigh, prec)
            update_labels(S, T)
            S_neigh = get_neighbors(S)

def hungarian_algorithm():
    for i in range(1, db_size):
        v = pick_free_block()
        S = {v}
        T = set()
        S_neigh = get_neighbors(S)
        prec = dict()
        # print(v, S, T, S_neigh, prec)

        hungarian_step(v, S, T, S_neigh, prec)
        cursor.execute("SELECT * FROM {}".format(matching_table))
        # print(cursor.fetchall())

            
initialize_labels()
initialize_equigraph()
initialize_matching()

hungarian_algorithm()

###### Imputation

# import numpy as np
# import pandas as pd
# from sklearn.impute import KNNImputer

# cursor.execute("SELECT * FROM {}".format(missing_data_table))
# tuples_list = cursor.fetchall()
# df = pd.DataFrame(tuples_list)

# imputer = KNNImputer(n_neighbors=10)
# impute = imputer.fit_transform(df)
# print(impute)
